"""Sample test script for package.

Performs unit tests on elements of the package.

Created: 2020-06-11 (Merijn, DAT-1583)
Updated: 2020-06-12 (Merijn, DAT-1583)
"""

# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
import json
import unittest

from vaknl_user import user
from vaknl_user import event


# ----------------------------------------------------------------------------------------------------------------------
# Test classes
# ----------------------------------------------------------------------------------------------------------------------
class UserTestCase(unittest.TestCase):

    def test_user_class_1(self):
        """Test creation of User object."""
        # Failure expected
        self.assertIsInstance(user.User(), user.User)

    def test_user_class_2(self):
        """Test creation of User object."""
        # Success expected
        dmp_user_id = 'xxx'
        self.assertIsInstance(user.User(dmp_user_id), user.User)

    def test_user_class_3(self):
        """Test creation of User object."""
        # Failure expected
        dmp_user_id = None
        self.assertIsInstance(user.User(dmp_user_id), user.User)

    def test_user_class_4(self):
        """Test creation of User object."""
        # Failure expected
        dmp_user_id = 12345
        self.assertIsInstance(user.User(dmp_user_id), user.User)

    # def test_user_class_5(self):
    #     for item in [User.User('xxx'), User.User()]:
    #         self.assertIsInstance(item, User.User)

    def test_create_user_from_clickstream(self):
        """Check if user can be created with clickstream data imported from Firestore."""
        dmp_user_id = u'0:jbf1sowm:G~wbWSdw7QLdVcGP6RVZ3GT0fVHeFuqc'
        u = user.User(dmp_user_id)
        u.create_user_from_clickstream()
        self.assertIn('0:k7s6g6jz:9T~qVFXlRTEHl0DUfFR5AKvF3Or4LY1U', u.statistics['dmp_session_ids'])


class EventTestCase(unittest.TestCase):

    def setUp(self):
        """Sets up test case environment. Repeats for every separate test."""
        with open('data/clickstream_dmp_session_sample.json') as json_file:
            data = json.load(json_file)
        self.clickstream_events = data.get('event_list')
        self.user = user.User('xxx')

    def tearDown(self):
        """Tears up test case environment. Repeats for every separate test."""
        self.clickstream_events = None
        self.user = None

    def test_event_class_1(self):
        e = self.clickstream_events[0]
        self.assertIsInstance(self.user.create_event(e), event.Event)


# ----------------------------------------------------------------------------------------------------------------------
# Command-line entry point
# ----------------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    unittest.main()



"""User statistics.

Contains dataclasses that keep track of several statistics when events are added.

Created: 2020-03-0? (Merijn)
Updated: 2020-07-14 (Merijn, DAT-1583)
"""

# TODO: improve section headers and structure.


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field, asdict
from typing import List, Dict, Any, Union
from datetime import datetime


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from vaknl_user import event
from vaknl_NBC import NBC


# ----------------------------------------------------------------------------------------------------------------------
# Statistic super dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Statistic:
    # TODO: docstring.

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        pass

    def validate(self):
        # TODO: docstring.
        pass


# ----------------------------------------------------------------------------------------------------------------------
# Field label information dataclasses
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class LabelInfo:
    # TODO: docstring.
    count: int = 0
    frequency: Dict[Any, int] = field(default_factory=dict)
    last: Union[str, int, None] = None

    def add_label(self, label: Union[str, int]):
        # TODO: docstring.
        self.last = label
        if label is not None:
            label = str(label)
            self.count += 1
            self.frequency[label] = self.frequency[label] + 1 if label in self.frequency else 1


@dataclass
class LabelListInfo:
    # TODO: docstring.
    frequency: Dict[Any, int] = field(default_factory=dict)

    def add_labels(self, labels: List[Union[str, int]]):
        # TODO: docstring.
        for label in labels:
            label = str(label)
            self.frequency[label] = self.frequency[label] + 1 if label in self.frequency else 1


# ----------------------------------------------------------------------------------------------------------------------
# Specific sub-sub dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class SearchPageSearchQueryStatistics(Statistic):
    # TODO: docstring.
    budget: LabelInfo = field(default_factory=LabelInfo)
    count: int = 0
    departure_airports: LabelListInfo = field(default_factory=LabelListInfo)
    departure_date: LabelInfo = field(default_factory=LabelInfo)
    departure_date_flex: LabelInfo = field(default_factory=LabelInfo)
    distance_to_beach: LabelInfo = field(default_factory=LabelInfo)
    durations: LabelListInfo = field(default_factory=LabelListInfo)
    geo: LabelListInfo = field(default_factory=LabelListInfo)
    meal_plans: LabelListInfo = field(default_factory=LabelListInfo)
    min_hotel_rating: LabelInfo = field(default_factory=LabelInfo)
    min_star_rating: LabelInfo = field(default_factory=LabelInfo)
    party_composition: LabelInfo = field(default_factory=LabelInfo)
    theme: LabelInfo = field(default_factory=LabelInfo)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        if isinstance(e, event.other.SearchPageSearchQuery):
            self.count += 1

            # Fields with LabelInfo.
            self.budget.add_label(e.budget)
            self.departure_date.add_label(e.departure_date)
            self.departure_date_flex.add_label(e.departure_date_flex)
            self.distance_to_beach.add_label(e.distance_to_beach)
            self.min_hotel_rating.add_label(e.min_hotel_rating)
            self.min_star_rating.add_label(e.min_star_rating)
            self.party_composition.add_label(e.party_composition)
            self.theme.add_label(e.theme)

            # Fields with LabelListInfo
            self.departure_airports.add_labels(e.departure_airports)
            self.durations.add_labels(e.durations)
            self.geo.add_labels(e.geo)
            self.meal_plans.add_labels(e.meal_plans)


# ----------------------------------------------------------------------------------------------------------------------
# Statistics sub dataclasses used in Statistics
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class GeneralStatistics(Statistic):
    # TODO: docstring.
    activity_first_timestamp: int = None
    activity_last_timestamp: int = None
    event_cnt: int = field(default=0)
    funnel_step: str = None  # = 'visitor'; the funnel names could change
    pageview_cnt: int = field(default=0)
    user_identifiers: Dict[str, List] = field(default_factory=dict)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        t = e.timestamp
        self.activity_first_timestamp = min(elem for elem in [self.activity_first_timestamp, t] if elem is not None)
        self.activity_last_timestamp = max(elem for elem in [self.activity_last_timestamp, t] if elem is not None)
        self.event_cnt += 1 if isinstance(e, event.event.Event) else 0
        self.pageview_cnt += 1 if isinstance(e, event.pageview.Pageview) else 0
        if isinstance(e, event.other.SfmcId):
            if e.email and e.email != 'undefined':
                self.user_identifiers['email'] = list(set(self.user_identifiers.get('email', []) + [e.email]))

    def validate(self):
        # TODO: docstring.
        # TODO: dynamic default check.
        return (
                self.activity_first_timestamp is not None
                and self.activity_last_timestamp is not None
                and self.funnel_step is not None
                and self.event_cnt > 0
        )


@dataclass
class OtherStatistics(Statistic):
    # TODO: docstring.
    filter_party_composition_freq: Dict = field(default_factory=dict)
    keuzehulp_show_top10_cnt: int = field(default=0)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        if isinstance(e, event.filter.GlobalFilterPartyComposition):
            self.filter_party_composition_freq[e.label] = self.filter_party_composition_freq[e.label] + 1 \
                if e.label in self.filter_party_composition_freq else 1
        self.keuzehulp_show_top10_cnt += 1 if isinstance(e, event.other.KeuzehulpShowTop10) else 0


@dataclass
class ProductStatistics(Statistic):
    # TODO: docstring.
    availability_check_cnt: int = field(default=0)
    filter_airport_cnt: int = field(default=0)
    filter_departure_date_cnt: int = field(default=0)
    filter_meal_plan_cnt: int = field(default=0)
    filter_flight_cnt: int = field(default=0)
    image_click_cnt: int = field(default=0)
    giata_id: LabelInfo = field(default_factory=LabelInfo)
    price_click_cnt: int = field(default=0)
    service_cnt: int = field(default=0)

    # NBC:
    country_code: LabelInfo = field(default_factory=LabelInfo)
    region_id: LabelInfo = field(default_factory=LabelInfo)
    city_id: LabelInfo = field(default_factory=LabelInfo)
    theme: LabelListInfo = field(default_factory=LabelListInfo)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        self.availability_check_cnt += 1 if isinstance(e, event.other.ProductAvailability) else 0
        self.filter_airport_cnt += 1 if isinstance(e, event.filter.ProductPageFilterAirport) else 0
        self.filter_departure_date_cnt += 1 if isinstance(e, event.filter.ProductPageFilterDepDate) else 0
        self.filter_meal_plan_cnt += 1 if isinstance(e, event.filter.ProductPageFilterMealPlan) else 0
        self.filter_flight_cnt += 1 if isinstance(e, event.filter.ProductPageFilterFlight) else 0
        self.image_click_cnt += 1 if isinstance(e, event.other.ImageClick) else 0
        self.price_click_cnt += 1 if isinstance(e, event.other.PriceClick) else 0
        self.service_cnt += 1 if isinstance(e, event.other.ProductService) else 0
        if isinstance(e, event.pageview.PageviewProduct):
            self.giata_id.add_label(e.giata_id)
            data = nbc_api.get_by_giata_id(int(e.giata_id))
            if data:
                self.country_code.add_label(data.get('country_code'))
                self.region_id.add_label(data.get('region_id'))
                self.city_id.add_label(data.get('city_id'))
                themes = [
                    key for key, value in data.items() if ('theme_' in key) and (key != 'theme_main') and int(value)
                ]
                self.theme.add_labels(themes)


@dataclass
class ReservationStatistics(Statistic):
    # TODO: docstring.
    booking: LabelInfo = field(default_factory=LabelInfo)
    extras_cnt: int = field(default=0)
    extras_select_extras_cnt: int = field(default=0)
    overview_cnt: int = field(default=0)
    pageview_cnt: int = field(default=0)
    personal_data_cnt: int = field(default=0)
    reservation_id_freq: List[str] = field(default_factory=list)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        if isinstance(e, event.reservation.ReservationBooked):
            if e.reservation_id not in self.reservation_id_freq and e.reservation_id is not None:
                self.reservation_id_freq.append(e.reservation_id)
                self.booking.add_label(e.giata_id)
        self.extras_cnt += 1 if isinstance(e, event.reservation.ReservationExtras) else 0
        self.extras_select_extras_cnt += 1 if isinstance(e, event.other.SelectExtrasBookingStep) else 0
        self.overview_cnt += 1 if isinstance(e, event.reservation.ReservationOverview) else 0
        self.pageview_cnt += 1 if isinstance(e, event.pageview.PageviewBookingStep) else 0
        self.personal_data_cnt += 1 if isinstance(e, event.reservation.ReservationPersonalData) else 0


@dataclass
class SearchStatistics(Statistic):
    # TODO: docstring.
    filter_select_cnt: int = field(default=0)
    pageview_cnt: int = field(default=0)
    search_query: SearchPageSearchQueryStatistics = field(default_factory=SearchPageSearchQueryStatistics)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        self.filter_select_cnt += 1 if isinstance(e, event.filter.SearchPageFilter) else 0
        self.pageview_cnt += 1 if isinstance(e, event.pageview.PageviewSearch) else 0
        self.search_query.add_event(e, nbc_api)


@dataclass
class SessionStatistics(Statistic):
    # TODO: docstring.
    count: int = field(default=0)
    day_of_week_freq: LabelInfo = field(default_factory=LabelInfo)
    dmp_session_id_list: List[str] = field(default_factory=list)
    hour_of_day_freq: LabelInfo = field(default_factory=LabelInfo)

    def add_event(self, e: event.event.Event, nbc_api: NBC):
        # TODO: docstring.
        self.dmp_session_id_list = list(set(self.dmp_session_id_list + [e.dmp_session_id]))
        self.count = len(self.dmp_session_id_list)
        if isinstance(e, event.other.Session):
            date_time = datetime.fromtimestamp(e.timestamp / 1e3)
            self.day_of_week_freq.add_label(date_time.weekday())  # Day of week as number: 0-6 (Monday-Sunday)
            self.hour_of_day_freq.add_label(date_time.hour)


# ----------------------------------------------------------------------------------------------------------------------
# Statistics dataclass used in User
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Statistics:
    # TODO: docstring.
    general: GeneralStatistics = field(default_factory=GeneralStatistics)
    other: OtherStatistics = field(default_factory=OtherStatistics)
    product: ProductStatistics = field(default_factory=ProductStatistics)
    reservation: ReservationStatistics = field(default_factory=ReservationStatistics)
    search: SearchStatistics = field(default_factory=SearchStatistics)
    session: SessionStatistics = field(default_factory=SessionStatistics)

    def add_event(self, e: event.event.Event, nbc_api: NBC = None, update_funnel_step=True):
        # TODO: docstring.
        if nbc_api is None:
            nbc_api = NBC()
        self.general.add_event(e, nbc_api)
        self.other.add_event(e, nbc_api)
        self.product.add_event(e, nbc_api)
        self.reservation.add_event(e, nbc_api)
        self.search.add_event(e, nbc_api)
        self.session.add_event(e, nbc_api)
        if update_funnel_step:
            self.update_funnel_step()

    def add_events(self, ee: event.event.Events, nbc_api: NBC = None):
        # TODO: docstring.
        ee.sort_by_timestamp_ascending()
        if nbc_api is None:
            nbc_api = NBC()
        for item in ee.event_list:
            self.add_event(item, nbc_api, update_funnel_step=False)
        self.update_funnel_step()

    def validate(self):
        # TODO: docstring.
        # TODO: include other checks.
        result = self.general.validate()
        return result

    def to_dict(self):
        """Get dictionary of fields. All nested dataclasses should be converted to dictionaries as well."""
        return asdict(self)

    def update_funnel_step(self):
        # TODO: docstring.

        # TODO: add ppcVisitorAcco, which is a funnel event.

        current_step = self.general.funnel_step
        if current_step != 'booked':
            if self.reservation.booking.count > 0:
                self.general.funnel_step = 'booked'
            elif current_step != 'in_market':
                if (
                        self.reservation.personal_data_cnt + self.reservation.overview_cnt
                        + self.reservation.extras_select_extras_cnt
                ) > 0:
                    self.general.funnel_step = 'in_market'
                elif current_step != 'active_plus':
                    if self.product.availability_check_cnt or self.product.filter_airport_cnt or \
                            self.product.filter_flight_cnt or self.product.filter_meal_plan_cnt or \
                            self.product.service_cnt or self.reservation.extras_cnt or \
                            self.product.image_click_cnt >= 5:
                        self.general.funnel_step = 'active_plus'
                    elif current_step != 'active':
                        if self.search.search_query.departure_date.count or self.product.filter_departure_date_cnt or \
                                bool(self.other.filter_party_composition_freq) or \
                                self.other.keuzehulp_show_top10_cnt or self.product.image_click_cnt >= 3 or \
                                self.search.filter_select_cnt >= 2:
                            self.general.funnel_step = 'active'
                        else:
                            self.general.funnel_step = 'visitor'



"""Tests for `vaknl_user/user/user.py`.

Created: 2020-07-14 (Merijn, DAT-1583)
Updated: 2020-07-14 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
import unittest


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from vaknl_NBC import NBC
from vaknl_user import event
from vaknl_user import user
from tests.utils import read_json


# ----------------------------------------------------------------------------------------------------------------------
# Test classes
# ----------------------------------------------------------------------------------------------------------------------
class UserMethodsTest(unittest.TestCase):
    """Test all the methods of the `user.user.User` dataclass."""

    # General attributes.
    nbc_api = NBC()
    firestore_client = user.utils.create_firestore_client(project_id='vaknldev')  # Don't test in prod environment!

    def setUp(self):
        """Sets up test case environment. Repeats for every separate test."""
        self.dmp_user_id = 'xxx'
        self.user = user.user.User(self.dmp_user_id)

    def tearDown(self):
        """Tears down test case environment. Repeats for every separate test."""
        self.dmp_user_id = None
        self.user = None

    def test_add_event(self):
        """Tests the `add_event()` method twice in a row."""
        # Add first event.
        n = self.user.statistics.general.event_cnt
        data = read_json('data/event/reservation/ReservationPersonalData.json')
        e = event.create_event(data)
        self.user.add_event(e, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + 1)

        # Add second event.
        data = read_json('data/event/filter/ProductPageFilterFlight.json')
        e = event.create_event(data)
        self.user.add_event(e, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + 2)

    def test_add_events(self):
        """Tests the `add_events()` method twice in a row."""
        # Add first events.
        n = self.user.statistics.general.event_cnt
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        ee = event.event.Events()
        ee.add_raw_events(data)
        m = len(ee.event_list)
        self.user.add_events(ee, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + m)

        # Add second events.
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        ee = event.event.Events()
        ee.add_raw_events(data)
        k = len(ee.event_list)
        self.user.add_events(ee, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + m + k)

    def test_add_raw_event(self):
        """Tests the `add_raw_event()` method twice in a row."""
        # Add first raw event.
        n = self.user.statistics.general.event_cnt
        data = read_json('data/event/reservation/ReservationPersonalData.json')
        self.user.add_raw_event(data, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + 1)

        # Add second raw event.
        data = read_json('data/event/filter/ProductPageFilterFlight.json')
        self.user.add_raw_event(data, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + 2)

    def test_add_raw_events(self):
        """Tests the `add_raw_events()` method twice in a row. Here, `data` is created from separate events, to avoid
        counting raw events in the raw data that are not converted to events later on."""
        # Add first raw events.
        n = self.user.statistics.general.event_cnt
        data_1 = read_json('data/event/pageview/PageviewProduct.json')
        data_2 = read_json('data/event/filter/ProductPageFilterAirport.json')
        data = [data_1, data_2]
        m = len(data)
        self.user.add_raw_events(data, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + m)

        # Add second raw events.
        data_1 = read_json('data/event/filter/GlobalFilterPartyComposition.json')
        data_2 = read_json('data/event/other/PriceClick.json')
        data_3 = read_json('data/event/reservation/ReservationExtras.json')
        data = [data_1, data_2, data_3]
        k = len(data)
        self.user.add_raw_events(data, self.nbc_api)
        self.assertIsInstance(self.user, user.user.User)
        self.assertEqual(self.user.statistics.general.event_cnt, n + m + k)

    def test_statistics_to_dict(self):
        """Test the `statistics_to_dict()` method."""
        # Check if dictionary.
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        self.user.add_raw_events(data, self.nbc_api)
        d = self.user.statistics_to_dict()
        self.assertIsInstance(d, dict)

        # Check if the fields in `Statistics` and the keys in the dictionary have a one-on-one relationship.
        fields = ['general', 'other', 'product', 'reservation', 'search', 'session']
        for field in fields:
            self.assertTrue(field in d)
        self.assertTrue(len([field for field in d.keys() if field not in fields]) == 0)

        # Add field that does not belong.
        self.user.statistics.not_a_field = 0
        d = self.user.statistics_to_dict()
        self.assertIsInstance(d, dict)
        self.assertTrue('not_a_field' not in d)

    def test_statistics_from_dict(self):
        """Test the `statistics_from_dict()` method."""
        # Is equal to original.
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        u = user.user.User(self.dmp_user_id)
        u.add_raw_events(data, self.nbc_api)
        d = u.statistics_to_dict()
        self.user.statistics_from_dict(d)
        self.assertEqual(self.user, u)

        # Add field that does not belong.
        d['not_a_field'] = 'I am not wanted.'
        self.user = user.user.User(self.dmp_user_id)
        self.user.statistics_from_dict(d)
        self.assertEqual(self.user, u)

    def test_import_statistics(self):
        """Test the `import_statistics()` method. For this test to work the `self.dmp_user_id` might have to be
        changed."""
        # The following `dmp_user_id` must exists in the `dmp_user` Firestore collection in order to import statistics.
        dmp_user_id = u'xxx'

        # Import statistics from Firestore.
        u = user.user.User(dmp_user_id)
        u.import_statistics(self.firestore_client)
        self.assertIsInstance(u, user.user.User)

    def test_export_statistics(self):
        """Test the `export_statistics()` method. For this test to work the `self.dmp_user_id` might have to be
        changed."""
        # The following `dmp_user_id` must not exists in the `dmp_user` Firestore collection in `vaknldev`, in order not
        # to overwrite existing data.
        dmp_user_id = u'dummy_id_that_should_not_exist'

        # Create user.
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        u = user.user.User(dmp_user_id)
        u.add_raw_events(data, self.nbc_api)

        # Write to Firestore.
        u.export_statistics(self.firestore_client)

        # Read from Firestore and create user to compare.
        v = user.user.User(dmp_user_id)
        v.import_statistics(self.firestore_client)
        self.assertEqual(u, v)

        # Delete written document from Firestore.
        user.utils.delete_dmp_user_firestore(self.firestore_client, dmp_user_id)

    def test_create_user_from_clickstream(self):
        """Test the `create_user_from_clickstream()` method. For this test to work the `dmp_user_id` variable might
        have to be changed."""
        # The following `dmp_user_id` must exists in the `in_website_clickstream` Firestore collection in order to
        # import data.
        dmp_user_id = u'0:ehwlstzz:4Bw6YDlX5g9XAUgpol_6AIgYPBOWcMYf'

        # Create user by using by using import function form utilities and the `add_raw_events()` method.
        u = user.user.User(dmp_user_id)
        data = user.utils.import_website_clickstream_events(self.firestore_client, u.dmp_user_id)
        self.assertTrue(len(data) > 0)  # Make sure there is at least one event imported.
        u.add_raw_events(data, self.nbc_api)
        self.assertIsInstance(u, user.user.User)

        # Create user with `create_user_from_clickstream()` method.
        v = user.user.User(dmp_user_id)
        v.create_user_from_clickstream(self.firestore_client)
        self.assertIsInstance(v, user.user.User)
        self.assertEqual(u, v)

    def test_update_user_from_clickstream(self):
        """Test the `create_user_from_clickstream()` method. For this test to work the `dmp_user_id` variable might
        have to be changed."""
        # The following `dmp_user_id` must exists in the `in_website_clickstream` Firestore collection in order to
        # import data.
        dmp_user_id = u'0:ehwlstzz:4Bw6YDlX5g9XAUgpol_6AIgYPBOWcMYf'

        # Create user by using by using import function form utilities and the `add_raw_events()` method.
        u = user.user.User(dmp_user_id)
        data = user.utils.import_website_clickstream_events(self.firestore_client, u.dmp_user_id)
        u.add_raw_events(data, self.nbc_api)
        self.assertIsInstance(u, user.user.User)

        # Make sure there are at least two events with different timestamp for this user.
        min_timestamp = u.statistics.general.activity_first_timestamp
        self.assertNotEqual(min_timestamp, u.statistics.general.activity_last_timestamp)

        # Create user that is partially filled with the first event.
        ee = event.event.Events()
        ee.add_raw_events(data)
        ee.event_list = [e for e in ee.event_list if e.timestamp <= min_timestamp]
        v = user.user.User(dmp_user_id)
        v.add_events(ee, self.nbc_api)
        self.assertGreater(v.statistics.general.event_cnt, 0)  # Make sure there is at least one event added.

        # Check if `update_user_from_clickstream()` method updates the user in the same way.
        v.update_user_from_clickstream(self.firestore_client)
        self.assertEqual(u, v)


# ----------------------------------------------------------------------------------------------------------------------
# Command-line entry point
# ----------------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    unittest.main()



"""User dataclass.

User class created from firestore website clickstream data.

Created: 2020-03-0? (Merijn)
Updated: 2020-07-14 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field, asdict
from typing import Dict, List
from dacite import from_dict


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from vaknl_user.user import utils
from vaknl_user.user.statistics import Statistics
from vaknl_user import event
from vaknl_NBC import NBC


# ----------------------------------------------------------------------------------------------------------------------
# User dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class User:
    """User dataclass for dmp_users."""
    dmp_user_id: str
    statistics: Statistics = field(default_factory=Statistics)

    def add_event(self, e: event.event.Event, nbc_api: NBC = None):
        """Adds single event from an `Event` object.

        Args:
            e: `Event` instance.
            nbc_api: NBC API instance, from which Non Bookable Content for an accommodation can be collected.
        """
        self.statistics.add_event(e, nbc_api)

    def add_events(self, ee: event.event.Events, nbc_api: NBC = None):
        """Adds multiple events represented as an `Events` object.

        Args:
            ee: `Events` instance that contains a list of `Event` instances.
            nbc_api: NBC API instance, from which Non Bookable Content for an accommodation can be collected.
        """
        self.statistics.add_events(ee, nbc_api)

    def add_raw_event(self, raw_event: dict, nbc_api: NBC = None):
        """Adds single raw clickstream event.

        Args:
            raw_event: Raw clickstream event.
            nbc_api: NBC API instance, from which Non Bookable Content for an accommodation can be collected.
        """
        e = event.create_event(raw_event)
        self.add_event(e, nbc_api)

    def add_raw_events(self, raw_events: List[Dict], nbc_api: NBC = None):
        """Adds multiple raw clickstream events.

        Args:
            raw_events: List of raw clickstream events.
            nbc_api: NBC API instance, from which Non Bookable Content for an accommodation can be collected.
        """
        ee = event.event.Events()
        ee.add_raw_events(raw_events)
        self.add_events(ee, nbc_api)

    def statistics_to_dict(self) -> dict:
        """Gets dictionary of fields in `statistics`. All nested dataclasses should be converted to dictionaries as
        well.

        Return:
            Dictionary representation of `statistics` field.
        """
        return self.statistics.to_dict()

    def statistics_from_dict(self, d: dict):
        """Fills `statistics` field by providing a dictionary representation. All nested dataclasses should be filled
        accordingly.

        Args:
            d: Dictionary representation of `statistics` field.
        """
        self.statistics = from_dict(data_class=Statistics, data=d)

    def import_statistics(self, client: utils.firestore.client):
        """Imports `statistics` field from Firestore, overwriting the field.

        Args:
            client: Firestore client instance.
        """
        d = utils.read_statistics_from_firestore(dmp_user_id=self.dmp_user_id, client=client)
        if d:
            self.statistics_from_dict(d)
        else:
            print(f'No `statistics` data found in the `{utils.USER_FIRESTORE_COLLECTION}` Firestore collection for '
                  f'`dmp_user_id`: `{self.dmp_user_id}`.')

    def export_statistics(self, client: utils.firestore.client):
        """Exports `statistics` field to Firestore, overwriting it.

        Args:
            client: Firestore client instance.
        """
        assert isinstance(self.statistics, Statistics), 'The `statistics` field need to be a `Statistics` instance.'
        assert self.statistics.validate(), \
            'User `statistics` are ill defined and could not be exported to Firestore. Tip: add at least one event ' \
            'to the user.'
        data = self.statistics.to_dict()
        utils.write_statistics_to_firestore(client=client, dmp_user_id=self.dmp_user_id, statistics=data)

    def create_user_from_clickstream(self, client: utils.firestore.client):
        """Gets all raw clickstream data and updates the user `statistics`.

        Args:
            client: Firestore client instance.
        """
        self.statistics = Statistics()  # Reset to default.
        data = utils.import_website_clickstream_events(client=client, dmp_user_id=self.dmp_user_id)
        self.add_raw_events(data)

    def update_user_from_clickstream(self, client: utils.firestore.client):
        """Gets all raw clickstream data that has occurred after the latest event tracked in the `statistics` and
        updates the user `statistics` with that data.

        Args:
            client: Firestore client instance.
        """
        data = utils.import_website_clickstream_events(
            client=client,
            dmp_user_id=self.dmp_user_id,
            timestamp_left_bound=self.statistics.general.activity_last_timestamp
        )
        self.add_raw_events(data)



"""Test `Events` dataclass.

Performs unit tests on `vaknl_user.event.event.Events` dataclass functionality.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-12 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
import unittest

from vaknl_user import event
from tests.utils import read_json


# ----------------------------------------------------------------------------------------------------------------------
# Test classes
# ----------------------------------------------------------------------------------------------------------------------
class EventsGeneralTestCase(unittest.TestCase):
    """Test case for general functionality of `Events` dataclass`."""

    def test_blank_creation(self):
        """Test creating a blank instance."""
        ee = event.event.Events()
        self.assertIsInstance(ee, event.event.Events)
        self.assertEqual(ee.event_list, [])

    def test_merge(self):
        """Test `merge()` method."""
        file_path = 'data/event/clickstream_event_list_sample_small.json'
        data = read_json(file_path)
        ee = event.event.Events()
        ee.add_raw_events(clickstream_event_list=data)
        n = len(ee.event_list)
        ee.merge(ee)
        self.assertIsInstance(ee, event.event.Events)
        self.assertEqual(len(ee.event_list), 2*n)


class AddEventTestCase(unittest.TestCase):
    """Test case to check if `self.add_event()` method of the `vaknl_user.event.event.Events` dataclass works properly.
    Gets data on events from `data/event/` directory."""

    def add_event_tester(self, file_path: str):
        data = read_json(file_path)
        ee = event.event.Events()
        for event_data in data:
            ee.add_raw_event(clickstream_event_data=event_data)
            self.assertIsInstance(ee, event.event.Events)

    def test_1(self):
        file_path = 'data/event/clickstream_event_list_sample_small.json'
        self.add_event_tester(file_path)

    def test_2(self):
        file_path = 'data/event/clickstream_event_list_sample_large.json'
        self.add_event_tester(file_path)

    def test_3(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_1.json'
        self.add_event_tester(file_path)

    def test_4(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_2.json'
        self.add_event_tester(file_path)


class AddEventsTestCase(unittest.TestCase):
    """Test case to check if `self.add_events()` method of the `vaknl_user.event.event.Events` dataclass works properly.
    Gets data on events from `data/event/` directory."""

    def add_events_tester(self, file_path: str):
        data = read_json(file_path)
        ee = event.event.Events()
        ee.add_raw_events(clickstream_event_list=data)
        self.assertIsInstance(ee, event.event.Events)

    def test_1(self):
        file_path = 'data/event/clickstream_event_list_sample_small.json'
        self.add_events_tester(file_path)

    def test_2(self):
        file_path = 'data/event/clickstream_event_list_sample_large.json'
        self.add_events_tester(file_path)

    def test_3(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_1.json'
        self.add_events_tester(file_path)

    def test_4(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_2.json'
        self.add_events_tester(file_path)


# ----------------------------------------------------------------------------------------------------------------------
# Command-line entry point
# ----------------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    unittest.main()



"""Event data classes.

Event dataclasses from website clickstream events.

Created: 2020-03-0? (Merijn)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, InitVar, field, asdict
from typing import List
import operator


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from vaknl_user import event


# ----------------------------------------------------------------------------------------------------------------------
# Event parent dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Event:
    """Event super class."""
    event_id: str
    timestamp: int
    dmp_session_id: str
    event_value: InitVar[dict] = None

    def __post_init__(self, event_value: dict):
        if event_value:
            # Fill event object with data from event_value_json. Function should be overwritten in child if used.
            self._fill_event_from_event_value(event_value)
        self._force_fields_not_none()

    def _fill_event_from_event_value(self, event_value: dict):
        """Dummy function. Should be overwritten in child dataclass when used."""
        pass

    def _raise_error_if_fields_are_none(self, fields: list):
        """Raises error if given fields are None."""
        is_none = [item for item in fields if getattr(self, item) is None]
        is_none_str = "['" + "', '".join(is_none) + "']"
        assert not is_none, f'The following fields should be declared: {is_none_str}.'

    def _force_fields_not_none(self):
        """Dummy function. Should be overwritten in child dataclass when used."""
        pass

    @staticmethod
    def _custom_factory_dict(data: dict) -> dict:
        """Remove `dummy` attribute in dict. The dummy field is used in otherwise empty parent dataclasses to avoid
        inheritance problems."""
        return dict(x for x in data if x[0] != "dummy")

    def to_dict(self):
        """Write attributes to dictionary."""
        return asdict(self, dict_factory=self._custom_factory_dict)


# ----------------------------------------------------------------------------------------------------------------------
# Events dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Events:
    """Class for multiple events together, with ordering method."""
    event_list: List[Event] = field(default_factory=list)

    def sort_by_timestamp_ascending(self):
        """Sorts events ascending by timestamp."""
        self.event_list = sorted(self.event_list, key=operator.attrgetter('timestamp'))

    def add_event(self, e: Event):
        """Add single event."""
        assert isinstance(e, Event), 'The provided `e` is not an `Event` instance.'
        self.event_list.append(e)

    def add_raw_event(self, clickstream_event_data: dict):
        """Add single event from data of a clickstream event."""
        e = event.create_event(clickstream_event_data)
        if e is not None:
            self.event_list.append(e)

    def add_raw_events(self, clickstream_event_list: list):
        """Add multiple events from list of clickstream events."""
        for item in clickstream_event_list:
            try:
                self.add_raw_event(item)
            except Exception as expt:
                # Broad except statement is necessary to avoid that an error for a single event breaks the entire
                # process.
                print(expt)

    def merge(self, ee):
        """Merge second Events object into self."""
        assert isinstance(ee, self.__class__), f'`ee` should be a {self.__class__} instance.'
        self.event_list += ee.event_list



import os
import pickle


class NBC(object):
    """Singleton NBC class

    This means there is always only one instance of this object. When the NBC class is created it will load the NBC data
    from the `NBC.pkl` file. This data file could be excluded from the package to avoid unrestricted information
    sharing. In such a case, add the `NBC.pkl` file at location.
    """
    __instance = None
    __nbc_data = None

    def __new__(cls):
        if NBC.__instance is None:
            NBC.__instance = object.__new__(cls)
            file_path = 'vaknl_NBC/NBC.pkl'  # Original location.
            if not os.path.exists(file_path):
                # When using a Cloud Function, the file should be situated in the root directory.
                file_path = 'NBC.pkl'
            with open(file_path, 'rb') as output:
                cls.__nbc_data = pickle.load(output)

        return NBC.__instance

    def get_by_giata_id(self, giata_id):
        return self.__nbc_data.get(giata_id)


if __name__ == '__main__':
    # Test cases.
    nbc = NBC()
    print(nbc.get_by_giata_id(18878))
    print(nbc.get_by_giata_id(123))



"""Test `create_event()`.

Performs unit tests on `vaknl_user.event.create_event()` function.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
import unittest
import os

from vaknl_user import event
from tests.utils import read_json


# ----------------------------------------------------------------------------------------------------------------------
# Test classes
# ----------------------------------------------------------------------------------------------------------------------
# class SearchPageSearchQueryTestCase(unittest.TestCase):
#
#     def test_default_filters(self):
#         with open('data/event/event_value_json_search_page_search_query_default_filters.json') as json_file:
#             self.event_value_json = json.load(json_file)
#
#         base_event = event.event.Event(event_id='fergw43gverv', timestamp=426452754, dmp_session_id='0:rgtw5wetrbf')
#         e = event._create_event_search_page_search_query(base_event, self.event_value_json)
#         self.assertIsInstance(e, event.other.SearchPageSearchQuery)
#
#     def test_custom_filters(self):
#         with open('data/event/event_value_json_search_page_search_query_custom_filters.json') as json_file:
#             self.event_value_json = json.load(json_file)
#
#         base_event = event.event.Event(event_id='fergw43gverv', timestamp=426452754, dmp_session_id='0:rgtw5wetrbf')
#         e = event._create_event_search_page_search_query(base_event, self.event_value_json)
#         self.assertIsInstance(e, event.other.SearchPageSearchQuery)


class CreateSpecificEventTestCase(unittest.TestCase):
    """Test case to check if `vaknl_user.event.create_event()` function creates specific events correctly. Gets data on
    events from `data/event/` directory.

    Note: The following events are missing from the data files:
        event.pageview.PageviewBlog
        event.pageview.PageviewDeal
        event.pageview.PageviewKeuzehulp
        event.pageview.PageviewOther
        event.other.KeuzehulpShowTop10
    """

    def event_category_tester(self, event_category: str):
        _module = __import__(f'vaknl_user.event.{event_category}', fromlist=['object'])
        dir_path = f'data/event/{event_category}'
        for file in os.listdir(dir_path):
            if file.endswith(".json"):
                file_path = os.path.join(dir_path, file)
                event_name = os.path.basename(file_path).replace('.json', '')
                event_dataclass = getattr(_module, event_name)
                event_data = read_json(file_path)
                e = event.create_event(data=event_data)
                self.assertIsInstance(e, event_dataclass)

    def test_filter(self):
        self.event_category_tester('filter')

    def test_other(self):
        self.event_category_tester('other')

    def test_pageview(self):
        self.event_category_tester('pageview')

    def test_reservation(self):
        self.event_category_tester('reservation')


class CreateEventsFromClickstreamTestCase(unittest.TestCase):
    """Test case to check if `vaknl_user.event.create_event()` function can create a list of events from a list of raw
    clickstream data. Gets data from `data/event/` directory."""

    def clickstream_events_tester(self, file_path: str):
        data = read_json(file_path)
        for event_data in data:
            e = event.create_event(event_data)
            if e is not None:
                self.assertIsInstance(e, event.event.Event)

    def test_1(self):
        file_path = 'data/event/clickstream_event_list_sample_small.json'
        self.clickstream_events_tester(file_path)

    def test_2(self):
        file_path = 'data/event/clickstream_event_list_sample_large.json'
        self.clickstream_events_tester(file_path)

    def test_3(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_1.json'
        self.clickstream_events_tester(file_path)

    def test_4(self):
        file_path = 'data/event/clickstream_event_list_sample_huge_2.json'
        self.clickstream_events_tester(file_path)


# ----------------------------------------------------------------------------------------------------------------------
# Command-line entry point
# ----------------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    unittest.main()



"""Pageview events.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field
import re


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from .event import Event
from vaknl_user.event.utils import safe_list_get


# ----------------------------------------------------------------------------------------------------------------------
# Pageview dataclasses
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Pageview(Event):
    """Pageview superclass.
    event_value_type == 'pageview'"""
    # url: str
    # page_type: str
    # The following (hidden) dummy is necessary to not create inheritance attribute order problems on defaults later on.
    dummy: int = field(repr=False, default=None)


# ----------------------------------------------------------------------------------------------------------------------
# Pageview sub-dataclasses
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class PageviewHome(Pageview):
    """pageview_page_type == 'homePage'"""
    funnel_event: bool = False


@dataclass
class PageviewSearch(Pageview):
    """pageview_page_type in ('brandedSearchPage', 'Search', 'nonBrandedSearchPage', 'Branded Search')
    Last 2 are old ones from before 2019."""
    funnel_event: bool = False


@dataclass
class PageviewProduct(Pageview):
    """pageview_page_type == 'productPage'"""
    giata_id: str = None
    funnel_event: bool = True

    def _fill_event_from_event_value(self, event_value: dict):
        self.giata_id = safe_list_get(re.findall(r"hotelId=(\d+)", event_value['pageview'].get('url')), 0, None)

    def _force_fields_not_none(self):
        fields = ['giata_id']
        self._raise_error_if_fields_are_none(fields)


@dataclass
class PageviewBookingStep(Pageview):
    """pageview_page_type == 'bookingForm'"""
    funnel_event: bool = False


@dataclass
class PageviewDeal(Pageview):
    """pageview_page_type == 'dealPage'"""
    funnel_event: bool = False


@dataclass
class PageviewKeuzehulp(Pageview):
    """pageview_page_type == 'searchAssistantPage'"""
    # page_type2: str
    funnel_event: bool = False


@dataclass
class PageviewContent(Pageview):
    """pageview_page_type == 'content'"""
    funnel_event: bool = False


@dataclass
class PageviewBlog(Pageview):
    """pageview_page_type == 'newsPage'"""
    funnel_event: bool = False


@dataclass
class PageviewError(Pageview):
    """pageview_page_type in ('errorPage', '404Page')"""
    funnel_event: bool = False


@dataclass
class PageviewOther(Pageview):
    """All other pageview events that are not defined as separate dataclass."""
    funnel_event: bool = False



"""Tests for `vaknl_user/user/statistics.py`.

Created: 2020-07-12 (Merijn, DAT-1583)
Updated: 2020-07-12 (Merijn, DAT-1583)
"""

# TODO: Unittests for all other class methods or functions.

# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
import unittest
from dacite import from_dict


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from vaknl_NBC import NBC
from vaknl_user import event
from vaknl_user import user
from tests.utils import read_json


# ----------------------------------------------------------------------------------------------------------------------
# Test classes
# ----------------------------------------------------------------------------------------------------------------------
class StatisticsMethodsTest(unittest.TestCase):

    nbc_api = NBC()
    base_data = read_json('data/event/other/Session.json')
    base_event = event.create_event(base_data)

    def setUp(self):
        """Sets up test case environment. Repeats for every separate test."""
        self.statistics = user.statistics.Statistics()
        self.statistics.add_event(self.base_event, self.nbc_api)

    def tearDown(self):
        """Tears down test case environment. Repeats for every separate test."""
        self.statistics = None

    def test_add_event(self):
        """Tests the `add_event()` method twice in a row."""
        # Add first event.
        n = self.statistics.general.event_cnt
        data = read_json('data/event/reservation/ReservationPersonalData.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api)
        self.assertIsInstance(self.statistics, user.statistics.Statistics)
        self.assertEqual(self.statistics.general.event_cnt, n + 1)

        # Add second event.
        data = read_json('data/event/filter/ProductPageFilterFlight.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api)
        self.assertIsInstance(self.statistics, user.statistics.Statistics)
        self.assertEqual(self.statistics.general.event_cnt, n + 2)

    def test_add_events(self):
        """Tests the `add_events()` method twice in a row."""
        # Add first events.
        n = self.statistics.general.event_cnt
        data = read_json('data/event/clickstream_event_list_sample_small.json')
        ee = event.event.Events()
        ee.add_raw_events(data)
        m = len(ee.event_list)
        self.statistics.add_events(ee, self.nbc_api)
        self.assertIsInstance(self.statistics, user.statistics.Statistics)
        self.assertEqual(self.statistics.general.event_cnt, n + m)

        # Add second events.
        data = read_json('data/event/clickstream_event_list_sample_large.json')
        ee = event.event.Events()
        ee.add_raw_events(data)
        k = len(ee.event_list)
        self.statistics.add_events(ee, self.nbc_api)
        self.assertIsInstance(self.statistics, user.statistics.Statistics)
        self.assertEqual(self.statistics.general.event_cnt, n + m + k)

    def test_to_dict(self):
        """Tests the `to_dict()` method."""
        d = self.statistics.to_dict()
        s = from_dict(data_class=user.statistics.Statistics, data=d)
        self.assertEqual(s, self.statistics)

    def test_validate(self):
        """Tests the `validate()` method."""
        # TODO: Test validate() method.
        pass

    def test_update_funnel_step(self):
        """Tests the `update_funnel_step()` method after adding several different funnel events."""
        # Test base user with only one `Session` event.
        self.assertEqual(self.statistics.general.funnel_step, 'visitor')

        # Test change to 'active'.
        data = read_json('data/event/filter/GlobalFilterPartyComposition.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api, update_funnel_step=False)
        self.assertEqual(self.statistics.general.funnel_step, 'visitor')
        self.statistics.update_funnel_step()
        self.assertEqual(self.statistics.general.funnel_step, 'active')

        # Test change to 'active_plus'.
        data = read_json('data/event/filter/ProductPageFilterFlight.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api, update_funnel_step=False)
        self.assertEqual(self.statistics.general.funnel_step, 'active')
        self.statistics.update_funnel_step()
        self.assertEqual(self.statistics.general.funnel_step, 'active_plus')

        # Test change to 'in_market'.
        data = read_json('data/event/reservation/ReservationPersonalData.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api, update_funnel_step=False)
        self.assertEqual(self.statistics.general.funnel_step, 'active_plus')
        self.statistics.update_funnel_step()
        self.assertEqual(self.statistics.general.funnel_step, 'in_market')

        # Test change to 'booked'.
        data = read_json('data/event/reservation/ReservationBooked.json')
        e = event.create_event(data)
        self.statistics.add_event(e, self.nbc_api, update_funnel_step=False)
        self.assertEqual(self.statistics.general.funnel_step, 'in_market')
        self.statistics.update_funnel_step()
        self.assertEqual(self.statistics.general.funnel_step, 'booked')


# ----------------------------------------------------------------------------------------------------------------------
# Command-line entry point
# ----------------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    unittest.main()
